//===-- temu-c/Loader.h - Binary image loading ------------------*- C++ -*-===//
//
// T-EMU: The Terma Emulator
// (c) Terma 2015
// Authors: Mattias Holm <maho (at) terma.com>
//
//===----------------------------------------------------------------------===//

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

/*! \brief Loads a binary image to the memory object.

  The function will autodetect the file type (based on extensions and
  magic headers in the file).

  Binary images will be placed at address 0.

  The image can be one of the supported file formats.

  \param mem Memory object, must conform to the mem interface.
  \param file Path to file to load.
  \result 0 on success, other values indicates errors.
 */
int temu_loadImage(void *mem, const char *file);

/*! \brief Loads a binary image to the memory object at the given
           address.

  The function will autodetect the file type (based on extensions and
  magic headers in the file). The binary is loaded to the given
  physical address (relocated). The relocation semantics is file type
  specific:

  ELF: The content is shifted by pa - start address. I.e. the start
       address ends up at pa.
  SREC: The content is loaded to pa - start address.
  BIN: The content is loaded to pa.

  \param mem Memory object, must conform to the mem interface.
  \param file Path to file to load.
  \param pa The physical address to where to load the file.
  \result 0 on success, other values indicates errors.
 */
int temu_loadRelocatedImage(void *mem, const char *file, uint64_t pa);

/*! \brief Loads a raw binary image to the memory object at the given
 *         address.
 *
 * The function will assume that the file is a raw binary. This way
 * you can load an ELF file as is to memory (e.g. where it is expected
 * by the boot loader).

  \param mem Memory object, must conform to the mem interface.
  \param file Path to file to load.
  \param pa Physical address where to load the image.
  \result 0 on success, other values indicates errors.
 */
int temu_loadBinaryImage(void *mem, const char *file, uint64_t pa);


int temu_loadElfImage(void *mem, const char *file, uint64_t pa);
int temu_loadSrecImage(void *mem, const char *file, uint64_t pa);

#ifdef __cplusplus
}
#endif
