//===-- temu-c/Objsys.h - T-EMU Object System -------------------*- C++ -*-===//
//
// T-EMU: The Terma Emulator
// (c) Terma 2015
// Authors: Mattias Holm <maho (at) terma.com>
//
//===----------------------------------------------------------------------===//

#ifndef TEMU_OBJSYS_C_H
#define TEMU_OBJSYS_C_H

#include <assert.h>
#include <stdint.h>
#include <stdlib.h>
#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __cplusplus
#define TEMU_PLUGIN_INIT \
  extern "C" void temu_pluginInit(void)
#else
#define TEMU_PLUGIN_INIT \
  void temu_pluginInit(void)
#endif

typedef void temu_Class;
typedef void temu_MetaIface;

typedef struct {
  temu_Class *Class;
} temu_Object;


// Generic interface
typedef struct temu_IfaceRef {
  void *Obj;
  void *Iface;
} temu_IfaceRef;

// In some cases, the number of objects we know about is unknown, so
// we need a dynamic array (e.g. std::vector in C++), but we cannot
// expose std::vector in a known format to LLVM, so we do this simple
// system with a dynamic interface array.
typedef struct temu_IfaceRefArray {
  uint32_t Size;
  uint32_t Reserved;
  temu_IfaceRef *Ifaces;
} temu_IfaceRefArray;

/*!
  Allocate interface array.

  \param Number of entries to reserve initially.
  \result The allocated interface array.
 */
temu_IfaceRefArray temu_ifaceRefArrayAlloc(unsigned Reserve);

/*!
  Push an object and raw interface pointer to an interface array.
  \param Arr The interface reference array.
  \param Obj The object implementing the relevant interface
  \param Iface Pointer to the interface struct.
 */
void temu_ifaceRefArrayPush2(temu_IfaceRefArray *Arr, void *Obj, void *Iface);
/*!
  Push an interface reference to the interface array.
  \param Arr The interface reference array
  \param Iface The interface reference (object-interface pointer pair)
 */
void temu_ifaceRefArrayPush(temu_IfaceRefArray *Arr, temu_IfaceRef Iface);

/*!
  Get the length of a dynamically allocated interface array.
  \param Arr The interface array to get the size from.
  \result The length in number of entries.
 */
unsigned temu_ifaceRefArraySize(temu_IfaceRefArray *Arr); // Return size

void temu_ifaceRefArrayDispose(temu_IfaceRefArray *Arr);


// Used to declare typed interfaces, also declares an array of such
// and inline wrappers for the iface array functions
#define TEMU_IFACE_REFERENCE_TYPE(N)                                    \
  typedef struct {                                                      \
    void *Obj;                                                          \
    N ## Iface *Iface;                                                  \
  } N ## IfaceRef;                                                      \
  typedef struct {                                                      \
    uint32_t Size;                                                      \
    uint32_t Reserved;                                                  \
    N ## IfaceRef *Ifaces;                                              \
  } N ## IfaceRefArray;                                                 \
  static inline N ## IfaceRefArray                                      \
  N ## IfaceRefArrayAlloc(unsigned Reserve)                             \
  {                                                                     \
    temu_IfaceRefArray Arr = temu_ifaceRefArrayAlloc(Reserve);          \
    N ## IfaceRefArray Res;                                             \
    Res.Size = Arr.Size;                                                \
    Res.Reserved = Arr.Reserved;                                        \
    Res.Ifaces = (N##IfaceRef*)Arr.Ifaces;                              \
    return Res;                                                         \
  }                                                                     \
  static inline void                                                    \
  N ## IfaceRefArrayDispose(N ## IfaceRefArray *Arr)                    \
  {                                                                     \
    temu_ifaceRefArrayDispose((temu_IfaceRefArray*)Arr);                \
  }                                                                     \
  static inline void                                                    \
  N ## IfaceRefArrayPush2(N ## IfaceRefArray *Arr, void *Obj, void *Iface) \
  {                                                                     \
    temu_ifaceRefArrayPush2((temu_IfaceRefArray*)Arr, Obj, Iface);      \
  }                                                                     \
  static inline void                                                    \
  N ## IfaceRefArrayPush(N ## IfaceRefArray *Arr, N ## IfaceRef Iface)  \
  {                                                                     \
    temu_IfaceRef Iface2;                                               \
    Iface2.Obj = Iface.Obj;                                             \
    Iface2.Iface = (void*)Iface.Iface;                                  \
                                                                        \
    temu_ifaceRefArrayPush((temu_IfaceRefArray*)Arr, Iface2);           \
  }                                                                     \
  static inline unsigned                                                \
  N ## IfaceRefArraySize(temu_IfaceRefArray *Arr)                       \
  {                                                                     \
    return Arr->Size;                                                   \
  }


#define TEMU_DYN_ARRAY_TYPE(T, P)                                       \
  typedef struct {                                                      \
    uint32_t Size;                                                      \
    uint32_t Reserved;                                                  \
    T *Values;                                                          \
  } temu_ ## P ## Array;                                                \
  static inline temu_ ## P ## Array                                     \
  temu_ ## P ## ArrayAlloc(unsigned Reserve)                            \
  {                                                                     \
    temu_ ## P ## Array Arr;                                            \
    Arr.Size = 0;                                                       \
    Arr.Reserved = Reserve;                                             \
    Arr.Values = (T*)calloc(Reserve, sizeof(T));                        \
    assert(Arr.Values);                                                 \
    return Arr;                                                         \
  }                                                                     \
  static inline void                                                    \
  temu_ ## P ## ArrayPush(temu_ ## P ## Array *Arr, T Val)              \
  {                                                                     \
    if (Arr->Reserved >= Arr->Size) {                                   \
      T *NewValues = (T*)realloc(Arr->Values, Arr->Reserved * 2);       \
      if (NewValues) {                                                  \
        Arr->Values = NewValues;                                        \
      } else {                                                          \
        abort();                                                        \
      }                                                                 \
    }                                                                   \
    Arr->Values[Arr->Size ++] = Val;                                    \
  }                                                                     \
  static inline unsigned                                                \
  temu_ ## P ## ArraySize(temu_ ## P ## Array *Arr)                     \
  {                                                                     \
    return Arr->Size;                                                   \
  }

TEMU_DYN_ARRAY_TYPE(int8_t, i8)
TEMU_DYN_ARRAY_TYPE(int16_t, i16)
TEMU_DYN_ARRAY_TYPE(int32_t, i32)
TEMU_DYN_ARRAY_TYPE(int64_t, i64)

TEMU_DYN_ARRAY_TYPE(uint8_t, u8)
TEMU_DYN_ARRAY_TYPE(uint16_t, u16)
TEMU_DYN_ARRAY_TYPE(uint32_t, u32)
TEMU_DYN_ARRAY_TYPE(uint64_t, u64)

typedef enum temu_Type {
  teTY_Invalid, // Invalid value 0

  // C pointer sized integers
  teTY_Intptr, // 1
  teTY_Uintptr, //2

  // Standard C floating point types
  teTY_Float, //3
  teTY_Double, //4

  // Standard C fixed width integer types
  teTY_U8, //5
  teTY_U16, //6
  teTY_U32, //7
  teTY_U64, //8
  teTY_I8,
  teTY_I16,
  teTY_I32,
  teTY_I64,

  // Object pointer, must be saved as an object reference
  teTY_Obj,

  // Internal pointer, points somewhere in the object itself
  // (e.g. bank resolution arrays) This can be saved as an offset...
  teTY_InternalPtr,

  // Interface references (object and interface pointer pair)
  teTY_IfaceRef,
  teTY_IfaceRefArray, // Dynamic object/interface array

  teTY_String, // C-string, useful for serialisation
} temu_Type;


/*!
  Typed property reference.

  The property reference contains a type tag and a raw pointer to the
  property.
 */
typedef struct temu_Propref {
  temu_Type Typ;
  void *Ptr;
} temu_Propref;

/*!
  Generic property value.

  As properties can be of any normal type, the propval struct provides
  a sum type/tagged union which contain both the type tag and the
  property value.
 */
typedef struct temu_Propval {
  temu_Type Typ;
  union {
    intptr_t IntPtr;
    uintptr_t UIntPtr;

    float f;
    double d;

    uint8_t u8;
    uint16_t u16;
    uint32_t u32;
    uint64_t u64;

    int8_t i8;
    int16_t i16;
    int32_t i32;
    int64_t i64;

    void *Obj;
    temu_IfaceRef IfaceRef;
    temu_IfaceRefArray IfaceRefArray;
    const char *String;
  };
} temu_Propval;


#ifdef PROP_ASSERTS_ENABLED
#define PROP_ASSERT(p, t) assert(p.Typ == t && "invalid property type")
#else
#define PROP_ASSERT(p, t)
#endif

// Ugly, but we want to be compatible with C++ and C99, meaning:
//   cannot use designated initializers (C99, not C++)
//   cannot use constructors (C++, not C99)

#define PROP_VAL_INITIALIZER(typ, suffix, typetag, valtag)        \
  static inline temu_Propval                                      \
  temu_makeProp ## suffix (typ val)                               \
  {                                                               \
    temu_Propval pv;                                              \
    pv.Typ = typetag;                                             \
    pv.valtag = val;                                              \
    return pv;                                                    \
  }                                                               \
  static inline typ                                               \
  temu_propValue ## suffix (temu_Propval pv)                      \
  {                                                               \
    PROP_ASSERT(pv.Typ, typetag);                                 \
    typ val = pv.valtag;                                          \
    return val;                                                   \
  }

PROP_VAL_INITIALIZER(intptr_t,  IntPtr,  teTY_Intptr,  IntPtr)
PROP_VAL_INITIALIZER(uintptr_t, UIntPtr, teTY_Uintptr, UIntPtr)

PROP_VAL_INITIALIZER(float,  Float,  teTY_Float,  f)
PROP_VAL_INITIALIZER(double, Double, teTY_Double, d)

PROP_VAL_INITIALIZER(uint8_t,  U8,  teTY_U8,  u8)
PROP_VAL_INITIALIZER(uint16_t, U16, teTY_U16, u16)
PROP_VAL_INITIALIZER(uint32_t, U32, teTY_U32, u32)
PROP_VAL_INITIALIZER(uint64_t, U64, teTY_U64, u64)

PROP_VAL_INITIALIZER(int8_t,   I8,  teTY_I8,  i8)
PROP_VAL_INITIALIZER(int16_t,  I16, teTY_I16, i16)
PROP_VAL_INITIALIZER(int32_t,  I32, teTY_I32, i32)
PROP_VAL_INITIALIZER(int64_t,  I64, teTY_I64, i64)

PROP_VAL_INITIALIZER(void*, Obj, teTY_Obj, Obj)
PROP_VAL_INITIALIZER(temu_IfaceRef, IfaceRef, teTY_IfaceRef, IfaceRef)

/*!  A prop writer function is provided when properties are
 *   registered, it is responsible for writing the property when
 *   temu_writeValue is called. The function can be used to introduce
 *   side-effects on writes unlike the setValue which will never have
 *   side-effects and only sets the raw value.
 */
typedef void (*temu_PropWriter)(void *Obj, temu_Propval Pv, int Idx);

/*!  A prop reader function is provided when properties are
 *   registered, it is responsible for reading the property when
 *   temu_readValue is called. The function can be used to introduce
 *   side-effects on reads unlike the getValue which will never have
 *   side-effects and only returns the raw value.
 */
typedef temu_Propval (*temu_PropReader)(void *Obj, int Idx);


/*!
  Get a reference to the named property.
  \param Obj An object of a class with the relevant property registered.
  \param PropName The name of the property to be queried.
 */
temu_Propref temu_getPropref(const void *Obj, const char *PropName);
int temu_getPropLength(const void *Obj, const char *PropName);
int temu_getPropDynLength(const void *Obj, const char *PropName);

int temu_isNumber(temu_Propval Pv);
int temu_isReal(temu_Propval Pv);
int temu_isDiscrete(temu_Propval Pv);
int temu_isString(temu_Propval Pv);

int64_t temu_asInteger(temu_Propval Pv);
uint64_t temu_asUnsigned(temu_Propval Pv);
double temu_asDouble(temu_Propval Pv);

/*
  Get and read value are different

  Get value has no side effects, while read value may have side effects.
 */
temu_Propval
temu_getValue(void *Obj, const char *PropName, int Idx);

// Python CTYPES does not like the anonymous union in propval.
uint8_t temu_getValueU8(void *Obj, const char *PropName, int Idx);
uint16_t temu_getValueU16(void *Obj, const char *PropName, int Idx);
uint32_t temu_getValueU32(void *Obj, const char *PropName, int Idx);
uint64_t temu_getValueU64(void *Obj, const char *PropName, int Idx);

int8_t temu_getValueI8(void *Obj, const char *PropName, int Idx);
int16_t temu_getValueI16(void *Obj, const char *PropName, int Idx);
int32_t temu_getValueI32(void *Obj, const char *PropName, int Idx);
int64_t temu_getValueI64(void *Obj, const char *PropName, int Idx);

/*!
  Read a property value with side-effects

  \param Obj Object pointer
  \param PropName Name of property to read.
  \param Idx Index of property, set to 0 if it is not an array.
  \result The property value corresponding to the name. In case of
          the property not being found, the Typ field of the returned
          property will be teTY_Invalid.
 */
temu_Propval
temu_readValue(void *Obj, const char *PropName, int Idx);

uint8_t  temu_readValueU8(void *Obj, const char *PropName, int Idx);
uint16_t temu_readValueU16(void *Obj, const char *PropName, int Idx);
uint32_t temu_readValueU32(void *Obj, const char *PropName, int Idx);
uint64_t temu_readValueU64(void *Obj, const char *PropName, int Idx);

int8_t  temu_readValueI8(void *Obj, const char *PropName, int Idx);
int16_t temu_readValueI16(void *Obj, const char *PropName, int Idx);
int32_t temu_readValueI32(void *Obj, const char *PropName, int Idx);
int64_t temu_readValueI64(void *Obj, const char *PropName, int Idx);

/*!
  Set a raw property value without side-effects

  \param Obj Object pointer
  \param PropName Name of property to read.
  \param Val The value to set. It must be of the correct type.
  \param Idx Index of property, set to 0 if it is not an array.
 */
void temu_setValue(void *Obj, const char *PropName, temu_Propval Val, int Idx);

void temu_setValueU8(void *Obj, const char *PropName, uint8_t Val, int Idx);
void temu_setValueU16(void *Obj, const char *PropName, uint16_t Val, int Idx);
void temu_setValueU32(void *Obj, const char *PropName, uint32_t Val, int Idx);
void temu_setValueU64(void *Obj, const char *PropName, uint64_t Val, int Idx);

void temu_setValueI8(void *Obj, const char *PropName, int8_t Val, int Idx);
void temu_setValueI16(void *Obj, const char *PropName, int16_t Val, int Idx);
void temu_setValueI32(void *Obj, const char *PropName, int32_t Val, int Idx);
void temu_setValueI64(void *Obj, const char *PropName, int64_t Val, int Idx);

/*!
  Write a property value with side-effects

  \param Obj Object pointer
  \param PropName Name of property to read.
  \param Val The value to set. It must be of the correct type.
  \param Idx Index of property, set to 0 if it is not an array.
 */
void temu_writeValue(void *Obj, const char *PropName, temu_Propval Val, int Idx);

void temu_writeValueU8(void *Obj, const char *PropName, uint8_t Val, int Idx);
void temu_writeValueU16(void *Obj, const char *PropName, uint16_t Val, int Idx);
void temu_writeValueU32(void *Obj, const char *PropName, uint32_t Val, int Idx);
void temu_writeValueU64(void *Obj, const char *PropName, uint64_t Val, int Idx);

void temu_writeValueI8(void *Obj, const char *PropName, int8_t Val, int Idx);
void temu_writeValueI16(void *Obj, const char *PropName, int16_t Val, int Idx);
void temu_writeValueI32(void *Obj, const char *PropName, int32_t Val, int Idx);
void temu_writeValueI64(void *Obj, const char *PropName, int64_t Val, int Idx);



temu_Propval temu_getNamedObjectProp(const char *Obj, const char *PropName,
                                     int Idx);

temu_Propval temu_readNamedObjectProp(const char *Obj, const char *PropName,
                                      int Idx);

void
temu_setNamedObjectProp(const char *Obj, const char *PropName,
                        temu_Propval Val, int Idx);

void
temu_writeNamedObjectProp(const char *Obj, const char *PropName,
                          temu_Propval Val, int Idx);



typedef struct temu_CreateArg {
  const char *Key;
  temu_Propval Val;
} temu_CreateArg;

#define TEMU_NULL_ARG { NULL, {teTY_Invalid} }

typedef void* (*temu_ObjectCreateFunc)(const char *Name,
                                       int Argc, const temu_CreateArg *Argv);
typedef void (*temu_ObjectDisposeFunc)(void*);

/*!
  Register a class in the TEMU object system. With these classes, the TEMU
  object system is responsible for allocation and de-allocation.

  \param ClsName The name of the class that is to be created.
  \param Create A constructor, the constructor is responsible for
                allocation and initialisation of the object.
  \param Dispose A destructor, this function is responsible for deleting
                 the objects of the class.
  \result A pointer to the class object that can be further used to register
          properties and interfaces.
 */
temu_Class*
temu_registerClass(const char *ClsName,
                   temu_ObjectCreateFunc Create,
                   temu_ObjectDisposeFunc Dispose);


/*!  Register an external class in the TEMU object system. An external
  class described objects that are managed outside the object
  system. Such classes can for example be used to bridge between two
  different object systems (e.g. TEMU and SMP2).

  \param ClsName The name of the class that is to be created.
  \result A pointer to the class object that can be further used to register
          properties and interfaces.
 */
temu_Class* temu_registerExternalClass(const char *ClsName);

#ifdef __cplusplus

/*!
  Add a property to a class.

  \param Cls The class object
  \param PropName Name of the property to add.

  \param Offset Offset in bytes from the start of the class to the
                value the property refers to.

  \param Typ Type of the property.

  \param Count Set to > 1 to indicate that the property is an array,
               the value is the length of the array.

  \param Wr Property write function. This function can have side-effects.
  \param Rd Property read function. This function can have side-effects.
*/
void
temu_addProperty(temu_Class *Cls, const char *PropName,
                 int Offset,
                 temu_Type Typ, int Count,
                 temu_PropWriter Wr = nullptr,
                 temu_PropReader Rd = nullptr,
                 const char *Doc = "");
#else
void
temu_addProperty(temu_Class *Cls, const char *PropName,
                 int Offset,
                 temu_Type Typ, int Count,
                 temu_PropWriter Wr,
                 temu_PropReader Rd,
                 const char *Doc);

#endif

/*!
 * Register interface with the named class.
 *
 * \param Cls_name Name of class to add interface to
 * \param Ifacename Interface name
 * \param IfaceType Interface type
 * \param Iface Pointer to the interface structure.
 * \param Count Reserved for future use.
                Set this to 0 or 1.
 */
#ifdef __cplusplus


void
temu_addInterface(temu_Class *Cls,
                  const char *IfaceName,
                  const char *IfaceType,
                  void *Iface, int Count = 0,
                  const char *Doc = "");

void*
temu_getInterface(void *Obj,
                  const char *IfaceName, int Idx = 0);

#else
void
temu_addInterface(temu_Class *Cls,

                  const char *IfaceName,
                  const char *IfaceType,
                  void *Iface, int Count,
                  const char *Doc);

void*
temu_getInterface(void *Obj,
                  const char *IfaceName, int Idx);

#endif

int temu_setVTable(temu_Class *Cls, void *VTable);
void* temu_getVTableForClass(temu_Class *Cls);
void* temu_getVTable(void *Obj);

int temu_isExternal(void *Obj);
int temu_isCpu(void *Obj);
int temu_isMachine(void *Obj);

/*! Erase all classes, interfaces, properties and objects registered
 *  in the object system. Not that objects of an external class will
 *  not be de-allocated. The responsibility of that lies with the
 *  programmer.
 */
void temu_objsysClear(void);

/*! Erase all objects, but do not delete classes.
 */
void temu_objsysClearObjects(void);

/*! Add object with external class.
 *
 * \param ClsName Name of class used for instantiation
 * \param ObjName Name of object in object system
 * \param Obj The object to add.
 * \result Obj if success, otherwise NULL.
 */
void* temu_addObject(const char *ClsName, const char *ObjName, void *Obj);


/*! Create object with class and name.
 *
 * \param ClsName Name of class used for instantiation
 * \param ObjName Name of object in object system
 *
 * \param Args Named argument pairs to pass into constructor. The list
 *             of args should be terminated by a
 *             TEMU_NULL_ARG. In case no arguments are passed,
 *             pass in NULL.
 *
 * \result Allocated object. Result is NULL in case of: class does not exist,
 *         object already allocated with the given name, out of memory.
 */
void* temu_createObject(const char *ClsName, const char *ObjName,
                        const temu_CreateArg *Args);



/*! Dispose object. The function will call the Objects dispose
 *  function.
 *
 * \param Obj The object to delete.
 */
void temu_disposeObject(void *Obj);

/*!
 * Get the class pointer for the named class
 */
temu_Class* temu_classForName(const char *ClsName);

/*!
 * Get the class for the given object.
 */
temu_Class* temu_classForObject(void *Obj);

/*!
 * Get class for named object
 */
temu_Class*
temu_classForObjectName(const char *Obj);


/*!
 * Get object for name
 */
void* temu_objectForName(const char *Name);

/*!
 * Get the name of an object
 */
const char* temu_nameForObject(const void *Obj);

/*!
 * Load a plugin. When a plugin is loaded the temu_pluginInit function
 * will be called. This function should create and define any classes
 * that the plugin provides.
 *
 * The function loads plugins following the operating system's loading
 * rules. This means in particular that first, the plugin will be
 * found in the RPATH location of libTEMUSupport (which refers to the
 * TEMU lib directory). Secondly, it will look in the LD_LIBRARY_PATH
 * and other standard load directories.
 *
 * \result 0 on success, non-zero otherwise.
 */
int temu_loadPlugin(const char *Path);

/*!
 * Get a string representing the type tag.
 */
const char* temu_typeToName(temu_Type Typ);


/*!

  Connect an interface property in object A to the interface in object
  B. If A.PropName is an interface array, B,B:IfaceName will be
  appended to it. For dynamic arrays, it is pushed to the end, for
  static arrays, it is placed in the first null slot.

  \param A The object to connect.
  \param PropName The name of the property in object A
  \param B The object to connect to.
  \param IfaceName The name of the interface in object B

 */
int temu_connect(void *A, const char *PropName, void *B, const char *IfaceName);

/*!
  Serialise the simulation state to the given file name.

  The serialisation writes a JSON formatted file and calls the
  serialise procedure in the (optional) ObjectIface if it is
  implemented.

  The serialisation interface can for example be used to write out
  memory buffers to separate files as such data should not be written
  to a JSON file.
*/
int temu_serialiseJSON(const char *FileName);


/*!
  De-serialise the simulation state.

  The function clears the whole internal object system, reads the JSON
  file FileName and creates all the objects.

  If a class implements the ObjectIface, the (optional) deserialise
  procedure will be called.
*/
int temu_deserialiseJSON(const char *FileName);

void temu_serialiseProp(void *Ctxt, const char *Name, temu_Type Typ,
                        int Count, void *Data);

void temu_deserialiseProp(void *Ctxt, void *Obj, const char *Name);

int temu_checkpointGetLength(void *Ctxt, const char *Name);
temu_Propval temu_checkpointGetValue(void *Ctxt, const char *Name, int Idx);

//! Check object system sanity.
//!
//! The function traverses all objects and their properties and ensures that
//! all (scalar) interface properties are connected.
//! An object can override the default check by implementing the checkSanity
//! function in the ObjectIface.
//!
//! \param Report set to 1 to have the function report connectivity issues on
//!        stdout, 2 to report on stderr, and 0 to not report at all.
//! \result Returns 0 if the object system is connected.
//!         Returns non-zero if the object system is not properly connected.
int temu_checkSanity(int Report);

/*!
 * Generic object interface The object interface provides generic
 * functionality such as serialisation and sanity checking support.
 */
typedef struct {
  void (*serialise)(void *Obj, const char *BaseName, void *Ctxt); // Optional
  void (*deserialise)(void *Obj, const char *BaseName, void *Ctxt); // Optional

  /*!
   * Return zero if the object is connected as expected, return
   * non-zero if the object is not fully connected the default check
   * will ensure that all the Interface references are connected, but
   * does not care about optional interfaces or
   */
  int (*checkSanity)(void *Obj, int Report); // Optional
} temu_ObjectIface;

TEMU_IFACE_REFERENCE_TYPE(temu_Object);

#ifdef __cplusplus
}
#endif


#endif /*!  TEMU_OBJSYS_C_H */
